from .gaussian_process import GaussianProcess as GP
from . import covariance_functions as cov
from . import mean_functions as mea
from . import likelihood_functions as lik
from . import inference_methods as inf
from . import fitting_methods as fit
from . import metrics as me
from . import data_manipulations as dm
from . import parameters as hp
from .plot import Plot as plot